@@ -20,7 +20,7 @@ use Locale::gettext;
 use Digest::MD5 qw(md5_hex);
 use HTML::FormEngine;
 @ISA = qw(HTML::FormEngine);
-$VERSION = '0.1';
+$VERSION = '0.2';
 
 ######################################################################
 
@@ -103,7 +103,9 @@ require HTML::FormEngine::DBSQL::Config;
 
 =head2 Example Database Table
 
-Execute the following (Postgre)SQL commands to create the table I used when developing the examples:
+Execute the following (Postgre)SQL commands to create the tables I used when developing the examples:
+
+    CREATE SEQUENCE user_uid_seq;
 
     CREATE TABLE "user" (
 	uid integer DEFAULT nextval('user_uid_seq'::text) NOT NULL,
@@ -125,11 +127,27 @@ Execute the following (Postgre)SQL commands to create the table I used when deve
 
     COMMENT ON COLUMN "user".email IS 'ERROR=rfc822;';
 
-    COMMENT ON COLUMN "user".phone IS 'ERROR_IN={{{not_null,digitonly},{not_null,digitonly}}};';
+    COMMENT ON COLUMN "user".phone IS 'ERROR_IN={{{not_null,digitonly},{not_null,digitonly}}};SUBTITLE={{,/}};SIZE={{5,10}};';
 
     COMMENT ON COLUMN "user".birthday IS 'ERROR=date;';
 
-Of course you can use any other table as well.
+    CREATE TABLE login (
+      uid integer DEFAULT currval('user_uid_seq'::text) NOT NULL,
+      username character varying(30) DEFAULT '-' NOT NULL,
+      "password" character varying(30) DEFAULT '-' NOT NULL
+    );
+
+    ALTER TABLE ONLY login
+    ADD CONSTRAINT login_pkey PRIMARY KEY (uid);
+
+    ALTER TABLE ONLY login
+    ADD CONSTRAINT "$1" FOREIGN KEY (uid) REFERENCES "user"(uid) MATCH FULL ON UPDATE CASCADE ON DELETE CASCADE;
+ 
+    COMMENT ON COLUMN login.password IS 'TYPE=password;VALUE=;ERROR=regex;regex=".{5,}";errmsg="must have more than 4 chars!";';
+
+    COMMENT ON COLUMN login.username IS 'ERROR=dbsql_unique;';
+
+Of course you can use any other table(s) as well.
 
 =head2 Example Output
 
@@ -360,7 +378,7 @@ entered data in a database.  Often very large forms are needed,
 e.g. when the user should provide his personal data to subscribe to an
 certain service.
 
-In most cases a SQL database is used. If you don't now anything about
+In most cases a SQL database is used. If you don't know anything about
 SQL databases or you're not using such things, this module will hardly
 help you. But if you do, you'll know that every record, that you want
 to store in a certain SQL database table, has to have certain fields
@@ -453,7 +471,7 @@ automatically passed to the I<not_null> check method. This means that
 their I<ERROR> variable will be set to I<not_null>.
 
 If the I<ERROR> variable was already set set through C<dbsql_preconf>,
-nothing will be changed.  If the variable was set through the fields
+nothing will be changed. If the variable was set through the fields
 comment (see below), the I<not_null> check will be added in front.
 
 If you called C<dbsql_set_show_default> and committed false (0), the
@@ -509,14 +527,20 @@ there as well:
 
 Note the ';' at the end of the trivial comment!
 
+In quoted areas ("..") '{', '}' and ',' are not interpreted. You can prevent the parsing of '"' and ';' by putting an '\' (backslash) in front.
+
 =head2 Methods
 
-=head3 dbsql_preconf ( HASHREF )
+=head3 dbsql_preconf ( HASHREF, PREPEND, APPEND )
 
-With this method, you can predefine some parts of the form
-configuration by hand.  The hash keys must be named after the tables
+In the referenced hash, you can predefine some parts of the form
+configuration by hand. The hash keys must be named after the tables
 fields. Every element must be a hash reference, in the referenced hash
 you can set variables.
+
+You can use the special keys I<prepend> and I<append> to add extra
+fields before or after the field.
+
 An example:
 
     my %preconf = (
@@ -531,15 +555,23 @@ An example:
 		   );
     $Form->dbsql_preconf(\%preconf);
 
+
+With PREPEND and APPEND you can define extra fields which will then be
+added to the top resp. the bottom of the generated form. If you want
+to add more than one field, you have to reference the array which
+contains the definitions, else you can reference the hash
+directly. See the POD of HTML::FormEngine for information about field
+definitions.
+
+B<Note:> If you pass more than one table name to C<dbsql_conf>, you
+must reference the fields with I<tablename.fieldname>!
+
 =cut
 
 ######################################################################
 
 sub dbsql_preconf {
-  my $self = shift;
-  my $preconf = shift;
-  my $prepend = shift;
-  my $append = shift;
+  my ($self,$preconf,$prepend,$append) = @_;
   if(ref($preconf) eq 'HASH') {
     $self->{dbsql_preconf} = merge($preconf, $self->{dbsql_preconf});
   }
@@ -549,23 +581,35 @@ sub dbsql_preconf {
 
 ######################################################################
 
-=head3 dbsql_conf ( TABLENAME, [ COUNT|WHERECONDITION, FIELDNAMES ] )
+=head3 dbsql_conf ( TABLENAME|ARRAYREF, [ COUNT|WHERECONDITION|HASHREF, FIELDNAMES|HASHREF ] )
 
 Creates an FormEngine-form-definition and calls FormEngines C<conf>
-method. If you provide COUNT, the form fields will be displayed COUNT
-times, which means that you can insert COUNT records.
+method.
+
+Normally you only want to manage records out of one table, then it is
+sufficient to give this tables name as first argument. But you can
+also pass several table names by using an array reference.
+
+If you provide COUNT, the form fields will be displayed COUNT times,
+which means that you can insert COUNT records.
 
 If you want to update records, you should provide WHERECONDITION
 instead. This must be a valid where-condition B<without> the 'WHERE'
-directive in front. DBSQL then shows input fields for every found
-record and uses the current values as defaults. The primary keys are
-stored in hidden fields, so that they can't be changed. Later they're
-used for updating the records.
+directive in front, or a hash reference. A hash reference you must
+provide if you passed several tablenames and want to define diffrent
+where conditions for theses tables. The keys must be the table names
+and the elements the appendant conditions.
+
+DBSQL then shows input fields for every found record and uses the
+current values as defaults. The primary keys are stored in hidden
+fields, so that they can't be changed. Later they're used for updating
+the records.
 
 If you'd like to set only some of the tables fields, put their names
 in an array and pass a reference to that as third and last argument
 (FIELDNAMES). If the first array element is '!', all fields which
-B<aren't> found in the array will be displayed.
+B<aren't> found in the array will be displayed. You must use a hash
+reference here, if you passed more than one table name.
 
 =cut
 
@@ -573,10 +617,10 @@ B<aren't> found in the array will be displayed.
 
 sub dbsql_conf {
   my $self = shift;
-  my (%donotuse, $sth, @pkey, @pkeyval, @fconf, @conf, @fields, $i, $sql, $count, $tmp, %value, $struct, $field);
+  my (%donotuse, $sth, @pkeyval, @fconf, @conf, @fields, $i, $sql, $count, $tmp, %value, $struct, $field, $tbl, $where, @pkey, %pkey);
 
-  $self->{dbsql_table} = shift || $self->{dbsql_table};
-  if(! defined($self->{dbsql_table}) || $self->{dbsql_table} eq '') {
+  $self->{dbsql_tables} = retarref(shift || $self->{dbsql_tables});
+  if(! defined($self->{dbsql_tables}) || ! @{$self->{dbsql_tables}}) {
     croak 'undefined table!';
   }
 
@@ -584,83 +628,129 @@ sub dbsql_conf {
 
   $self->{dbsql_fields} = shift || $self->{dbsql_fields};
 
-  $self->{dbsql_fields} = retarref($self->{dbsql_fields});
-
-  if($self->{dbsql_fields}->[0] eq '!') {
-    delete $self->{dbsql_fields}->[0];
-    foreach $_ (@{$self->{dbsql_fields}}) {
-      $donotuse{$_} = 1;
+  if(@{$self->{dbsql_tables}} > 1 && ref($self->{dbsql_fields}) ne 'HASH') {
+    croak 'fields must be assigned to tables!';
+  }
+  elsif(@{$self->{dbsql_tables}} == 1 && ref($self->{dbsql_fields}) ne 'HASH') {
+    $self->{dbsql_fields} = {$self->{dbsql_tables}->[0] => retarref($self->{dbsql_fields})};
+  }
+  if(ref($self->{dbsql_pkey}) ne 'HASH') {
+    $_ = retarref($self->{dbsql_pkey});
+    $self->{dbsql_pkey} = {$self->{dbsql_tables}->[0] => {}};
+    foreach $_ (@{$_}) {
+      $self->{dbsql_pkey}->{$self->{dbsql_tables}->[0]}->{$_} = 1;
     }
-    $self->{dbsql_fields} = ['%']
   }
-
-  if(! @{$self->{dbsql_fields}}) {
-    croak 'no fields defined!';
-  }
-
-  @pkey = $self->{dbsql}->primary_key(undef, undef, $self->{dbsql_table});
-
-  foreach $_(@pkey) {
-    $self->{dbsql_pkey}->{$_} = 1;
-  }
-
-  if($self->{dbsql_where} =~ m/^[0-9]+$/) {
+  if(! ref($self->{dbsql_where}) and $self->{dbsql_where} =~ m/^[0-9]+$/) {
     $count = $self->{dbsql_where};
     $self->{dbsql_show_default} = 1 if($self->{dbsql_show_default} == -254);
   }
-
+  elsif(ref($self->{dbsql_where}) ne 'HASH') {
+    $where = $self->{dbsql_where};
+    $self->{dbsql_where} = {};
+  } else {
+    $where = '';
+  }
+  foreach $tbl (@{$self->{dbsql_tables}}) {
+    if(ref($self->{dbsql_fields}->{$tbl}) ne 'ARRAY' or ! @{$self->{dbsql_fields}->{$tbl}}) {
+      $self->{dbsql_fields}->{$tbl} = ['%'];
+    }
+    $self->{dbsql_where}->{$tbl} = $where if(ref($self->{dbsql_where}) eq 'HASH' && ! defined($self->{dbsql_where}->{$tbl}));
+    $donotuse{$tbl} = {};
+    if($self->{dbsql_fields}->{$tbl}->[0] eq '!') {
+      delete $self->{dbsql_fields}->{$tbl}->[0];
+      foreach $_ (@{$self->{dbsql_fields}->{$tbl}}) {
+	$donotuse{$tbl}->{$_} = 1;
+      }
+      $self->{dbsql_fields}->{$tbl} = ['%']
+    }
+    if(! defined($self->{dbsql_pkey}->{$tbl})) {
+      $self->{dbsql_pkey}->{$tbl} = {};
+      foreach $_ ($self->{dbsql}->primary_key(undef, undef, $tbl)) {
+    	$field = (@{$self->{dbsql_tables}} > 1 ? "$tbl.$_" : $_);
+    	$self->{dbsql_pkey}->{$tbl}->{$field} = 1;
+      }
+    }
+  }
 
   if(defined($self->{dbsql_prepend})) {
     push @fconf, @{retarref($self->{dbsql_prepend})};
   }
 
-  $struct = $self->get_tbl_struct(\%donotuse);
-
-  foreach $tmp (@{$struct}) {
-    push @fields, $tmp->{name};
-    $_ = $self->_dbsql_makeconf($tmp);
-    if(defined($_->{prepend})) {
-      push @fconf, @{retarref($_->{prepend})};
-      delete $_->{prepend};
-    }
-    push @fconf, $_;
-    if(defined($_->{append})) {
-      push @fconf, @{retarref($_->{append})};
-      delete $_->{append};
+  foreach $tbl (@{$self->{dbsql_tables}}) {
+    $struct = $self->get_tbl_struct($tbl,$self->{dbsql_fields}->{$tbl},$donotuse{$tbl});
+    $self->{dbsql_fields}->{$tbl} = [];
+    foreach $tmp (@{$struct}) {
+      $_ = $self->_dbsql_makeconf($tmp,$tbl);
+      push @{$self->{dbsql_fields}->{$tbl}}, $_->{fname};
+      if(defined($_->{prepend})) {
+	push @fconf, @{retarref($_->{prepend})};
+	delete $_->{prepend};
+      }
+      push @fconf, $_;
+      if(defined($_->{append})) {
+	push @fconf, @{retarref($_->{append})};
+	delete $_->{append};
+      }
     }
   }
 
-  $self->{dbsql_fields} = \@fields;
-
   if(defined($self->{dbsql_append})) {
     push @fconf, @{retarref($self->{dbsql_append})};
   }
 
-  
-  if(! defined($count)) {
-    $sql = 'SELECT '.join(', ', @{$self->{dbsql_fields}}).' FROM "'.$self->{dbsql_table}.'"';
-    if($self->{dbsql_where} ne '') {
-      $sql .= ' WHERE '.$self->{dbsql_where};
+  foreach $tbl (@{$self->{dbsql_tables}}) {
+    foreach $field (keys(%{$self->{dbsql_pkey}->{$tbl}})) {
+      delete $self->{dbsql_pkey}->{$tbl}->{$field} unless(grep {$field eq $_} @{$self->{dbsql_fields}->{$tbl}});
     }
-    $sth = $self->{dbsql}->prepare($sql);
-    $sth->execute;
-    $count = $sth->rows;
-    while($self->{dbsql_show_value} && ($tmp = $sth->fetchrow_hashref)) {
-      foreach $_ (keys(%{$tmp})) {
-	if(ref($value{$_}) ne 'ARRAY') {
-	  $value{$_} = [];
-	}
-	if(defined($tmp->{$_}) and $tmp->{$_} =~ m/^({.*})$/) {
-	  push @{$value{$_}}, $self->_dbsql_parse($1);
+  }
+
+  if(! defined($count)) {
+    $count = -254;
+    foreach $tbl (@{$self->{dbsql_tables}}) {
+      $sql = 'SELECT ';
+      foreach $_ (@{$self->{dbsql_fields}->{$tbl}}) {
+	if(m/^(.+)?\.(.+)/) {
+	  $sql .= $self->{dbsql}->quote_identifier($1) . '.' . $self->{dbsql}->quote_identifier($2);
 	}
 	else {
-	  push @{$value{$_}}, $tmp->{$_};
+	  $sql .= $self->{dbsql}->quote_identifier($_);
+	}
+	$sql .= ',';
+      }
+      $sql =~ s/,$//;
+      $sql .= ' FROM ' . $self->{dbsql}->quote_identifier($tbl);
+      if($self->{dbsql_where}->{$tbl} ne '') {
+	$sql .= ' WHERE '.$self->{dbsql_where}->{$tbl};
+      }
+      $sth = $self->{dbsql}->prepare($sql);
+      if(! $sth->execute) {
+	$self->_dbsql_sql_error($sth->{Statement});
+      }
+      if($count == -254) {
+	$count = $sth->rows 
+      }
+      elsif($count ne $sth->rows) {
+	croak('There must be the same count of records for each table!');
+      }
+      while($self->{dbsql_show_value} && ($tmp = $sth->fetchrow_hashref)) {
+	foreach $_ (keys(%{$tmp})) {
+	  $field = (@{$self->{dbsql_tables}} > 1 ? "$tbl.$_" : $_);
+	  if(ref($value{$field}) ne 'ARRAY') {
+	    $value{$field} = [];
+	  }
+	  if(defined($tmp->{$_}) and $tmp->{$_} =~ m/^({.*})$/) {
+	    push @{$value{$field}}, ($self->_dbsql_parse($1));
+	  }
+	  else {
+	    push @{$value{$field}}, $tmp->{$_};
+	  }
 	}
       }
     }
   }
 
-  if($count > 1 and $self->{dbsql_row} == -254) {
+  if($self->{dbsql_row} > 0 or $count > 1 and $self->{dbsql_row} == -254) {
     $self->{dbsql_row} = 1;
     $tmp = [];
     foreach $_ (@fconf) {
@@ -670,6 +760,11 @@ sub dbsql_conf {
     push @conf, {templ => 'title', TITLE => $tmp};
   }
 
+  @pkey = ();
+  foreach $_ (keys(%{$self->{dbsql_pkey}})) {
+    push @pkey, keys(%{$self->{dbsql_pkey}->{$_}});
+  }
+
   for($i=0; $i<$count; $i++) {
     @pkeyval = ();
     $tmp = clone(\@fconf);
@@ -677,14 +772,14 @@ sub dbsql_conf {
       foreach $field (@{$tmp}) {
 	if(keys(%value)) {
 	  if(defined($field->{fname}) && defined($value{$field->{fname}})) {
-	    $field->{VALUE} = shift @{$value{$field->{fname}}};
+	    $field->{VALUE} = shift @{$value{$field->{fname}}} unless(defined($field->{VALUE}));
 	  }
 	}
-	push @pkeyval, $field->{VALUE} if ($self->{dbsql_secret} && grep {$_ eq $field->{fname}} @pkey);
+	push @pkeyval, $field->{VALUE} if ($field->{fname} and $self->{dbsql_secret} && grep {$_ eq $field->{fname}} @pkey);
       }
     }
 
-    $_ = md5_hex(join($self->{dbsql_secret}, @pkeyval));
+    $_ = md5_hex(join($self->{dbsql_secret}, @pkeyval) . $self->{dbsql_secret});
     push @{$tmp}, {templ => 'hidden', NAME => 'md5hash', VALUE => $_} if(@pkeyval);
     if($self->{dbsql_row} > 0) {
       push @conf, {templ => $self->{dbsql_row_tmpl}, ROWNUM => $i+1, sub => $tmp};
@@ -695,7 +790,6 @@ sub dbsql_conf {
   }
   
   $self->conf([{templ => 'body', sub => \@conf}]);
-  $self->add_default({default => {'FCOUNT' => scalar @fields}});
   if($self->{debug}) {
     foreach $_ (@fconf) {
       print $_->{NAME}, "\n";
@@ -723,14 +817,22 @@ for more information.
 
 sub dbsql_update {
   my ($self) = @_;
-  my ($md5hash, @pkeyval, @pkeyval2, $ok, $val);
+  my ($md5hash, @pkeyval, @pkeyval2, $ok, $val, $tbl);
+  $_ = 0;
+  foreach $tbl (@{$self->{dbsql_tables}}) {
+    $_ = $tbl and last unless(keys(%{$self->{dbsql_pkey}->{$tbl}}));
+  }
+  if($_) {
+    $self->_add_to_output($self->{dbsql_errmsg_tmpl}, {ERRMSG => gettext('Primary key is missing for table') . ' \'' . $_ . '\'!'});
+    return 0;
+  }
   if(! $self->{dbsql_hide_pkey} or $md5hash = $self->get_input_value('md5hash')) {
-
     if($self->{dbsql_hide_pkey}) {
-      foreach $_ (keys(%{$self->{dbsql_pkey}})) {
-	push @pkeyval, $self->get_input_value($_);
+      foreach $tbl (@{$self->{dbsql_tables}}) {
+	foreach $_ (keys(%{$self->{dbsql_pkey}->{$tbl}})) {
+	  push @pkeyval, $self->get_input_value($_);
+	}
       }
-      
       if(ref($md5hash) eq 'ARRAY') {
 	$ok = 1;
 	foreach $_ (@{$md5hash}) {
@@ -748,7 +850,7 @@ sub dbsql_update {
     }
 
     return $self->_dbsql_write(1) if($ok);
-
+    $self->_add_to_output($self->{dbsql_errmsg_tmpl}, {ERRMSG => gettext('Can\'t update record(s) due to primary key cheksum mismatch').'!'});
     return 0;
   }
 
@@ -770,7 +872,6 @@ valid (see FormEngines C<ok> method).
 
 =cut
 
-
 ######################################################################
 
 sub dbsql_insert {
@@ -844,7 +945,7 @@ sub dbsql_set_show_value {
 
 ######################################################################
 
-=head3 dbsql_set_pkey ( SCALAR|ARRAYREF )
+=head3 dbsql_set_pkey ( SCALAR|ARRAYREF|HASHREF )
 
 Normally the primary key of an database table is
 autodetected. Sometimes someone might like to define other fields as
@@ -853,16 +954,30 @@ records). You can pass a fieldname or an reference to an array with
 fieldnames to this method. This method should be called before
 C<dbsql_conf> (for being sure, call this method as early as possible).
 
+B<Note>: If you pass several table names to dbsql_conf, you must pass as hash reference here, else the passed pkeys will only be used for the first table.
+
 =cut
 
 ######################################################################
 
 sub dbsql_set_pkey {
-  my $self = shift;
-  my $pkey = retarref(shift);
-  foreach $_ (@{$pkey}) {
-    $self->{dbsql_pkey}->{$_} = 1;
+  my ($self,$pkey)= @_;
+  my $tbl;
+  if($pkey) {
+    if(ref($pkey) ne 'HASH') {
+      $self->{dbsql_pkey} = $pkey;
+      return 1;
+    }
+    foreach $tbl (keys(%{$pkey})) {
+      $self->{dbsql_pkey}->{$tbl} = {} if(ref($self->{dbsql_pkey}->{$tbl}) ne 'HASH');
+      $pkey->{$tbl} = [$pkey->{$tbl}] if(ref($pkey->{$tbl}) ne 'ARRAY');
+      foreach $_ (@{$pkey->{$tbl}}) {
+	$self->{dbsql_pkey}->{$tbl}->{"$tbl.$_"} = 1;
+      }
+    }
+    return 1;
   }
+  return 0;
 }
 
 ######################################################################
@@ -967,13 +1082,31 @@ sub dbsql_set_sqlerr_templ {
 =head3 dbsql_set_secret ( SECRET )
 
 If you want to update records, you can use the C<dbsql_update> method.
-That method uses the given values of the primary key to create where conditions, so that the right records are updated. The weak point is, that someone could corrupt the input data, so that the primary key values are changed and the wrong records are updated. To prevent this, for every record a extra hidden field is created which contains the md5 sum of the primary key concatenated with a secret string. So it is recognized if a primary key value was changed (because the newly created md5 sum won't match the submitted md5 sum).
+That method uses the given values of the primary key to create where
+conditions, so that the right records are updated. The weak point is,
+that someone could corrupt the input data, so that the primary key
+values are changed and the wrong records are updated. To prevent this,
+for every record a extra hidden field is created which contains the
+md5 sum of the primary key concatenated with a secret string. So it is
+recognized if a primary key value was changed (because the newly
+created md5 sum won't match the submitted md5 sum).
 
-With this method you can set the secret string. By default it is set to NULL, which means that calling C<dbsql_conf> will raise an error. For security reason an update isn't allowed without a secret string, except you pass false (0) to C<dbsql_set_hide_pkey>, which will allow changing the primary key and so no secret string will be needed.
+With this method you can set the secret string. By default it is set
+to NULL, which means that calling C<dbsql_conf> will raise an
+error. For security reason an update isn't allowed without a secret
+string, except you pass false (0) to C<dbsql_set_hide_pkey>, which
+will allow changing the primary key and so no secret string will be
+needed.
 
-Another possibilty is changing the value of C<$secret> in I<Config.pm> and so set a valid default secret string. But be careful, someone might just edit Config.pm and so get the secret string, whereas using diffrent keys in your scripts is much more secure.
+Another possibilty is changing the value of C<$secret> in I<Config.pm>
+and so set a valid default secret string. But be careful, someone
+might just edit Config.pm and so get the secret string, whereas using
+diffrent keys in your scripts is much more secure.
 
-It is recommended that you set the read permissions of scripts, which define secret keys, as restrictive as possible. For cgi scripts this means, that only the webserver user (mostly I<nobody> oder I<www-data>) must be able to read them.
+It is recommended that you set the read permissions of scripts, which
+define secret keys, as restrictive as possible. For cgi scripts this
+means, that only the webserver user (mostly I<nobody> oder
+I<www-data>) must be able to read them.
 
 =cut
 
@@ -1057,6 +1190,29 @@ sub dbsql_get_sqlerr {
   return @{$self->{dbsql_sqlerr}};
 }
 
+######################################################################
+
+=head3 dbsql_add_extra_sql(SQLCOMMAND, ARRAY)
+
+This method can be used to define some more sql commands which then
+will be executed for each record when C<insert> or <update> is called.
+
+The sql command might contain '?' (question marks). These will be
+replaced with the current values of the fields defined by the second
+argument. The first '?' is replaced with the value of the first
+element and so on.
+
+A backslash before an question mark will prevent it from being parsed.
+
+=cut
+
+######################################################################
+
+sub dbsql_add_extra_sql {
+  my($self,$sql,@vars) = @_;
+  push @{$self->{dbsql_extra_sql}}, [$sql, @vars] if($sql);
+}
+
 ######################################################################
 # INTERNAL METHODS                                                   #
 ######################################################################
@@ -1067,7 +1223,8 @@ sub _initialize_child {
   $self->{dbsql_preconf} = {};
   $self->{dbsql_where} = 1;
   $self->{dbsql_pkey} = {};
-  $self->{dbsql_fields} = ['%'];
+  $self->{dbsql_tables} = [];
+  $self->{dbsql_fields} = {};
   $self->{dbsql_hide_pkey} = 1;
   $self->{dbsql_show_value} = 1;
   $self->{dbsql_show_default} = -254;
@@ -1079,9 +1236,19 @@ sub _initialize_child {
   $self->{dbsql_empty_tmpl} = 'empty';
   $self->{dbsql_secret} = $HTML::FormEngine::DBSQL::Config::secret;
   $self->{dbsql_row} = -254;
+  $self->{dbsql_extra_sql} = [];
   foreach $_ (keys(%HTML::FormEngine::DBSQL::Config::skin)) {
     $self->{skins_av}->{$_} = $HTML::FormEngine::DBSQL::Config::skin{$_};
   }
+  foreach $_ (keys(%HTML::FormEngine::DBSQL::Config::checks)) {
+    $self->{checks}->{$_} = $HTML::FormEngine::DBSQL::Config::checks{$_};
+  }
+  foreach $_ (keys(%HTML::FormEngine::DBSQL::Config::confirm_skin)) {
+    $self->{confirm_skin}->{$_} = $HTML::FormEngine::DBSQL::Config::confirm_skin{$_};
+  }
+  foreach $_ (keys(%HTML::FormEngine::DBSQL::Config::confirm_handler)) {
+    $self->{confirm_handler}->{$_} = $HTML::FormEngine::DBSQL::Config::confirm_handler{$_};
+  }
   $self->set_skin('DBSQL');
 
   $self->_dbsql_initialize_child;
@@ -1091,72 +1258,76 @@ sub _dbsql_write {
   my $self = shift;
   my $update = shift;
   my %input = %{clone($self->{input})};
-  my ($sql, $value, @fields, @tbfields, @values, $i, %pkey, $tmp);
+  my ($sql, $value, $tbl, %fields, %values, @sql, $i, $x, %pkey, $key);
   my $while = [];
   my $res = 0;
-  
-  $tmp = {};
-  foreach $_ (@{$self->{dbsql_fields}}) { $tmp->{$_} = 1; }
-  
+    
   $self->_dbsql_sum_arr_elems(\%input);
   
-  foreach $_ (keys(%input)) {
-    if(defined($input{$_}) && ($_ =~ m/^[^\]\[]+$/) && ($tmp->{$_})) {
-      if(ref($input{$_}) ne 'ARRAY') {
-	$input{$_} = [$input{$_}]
-      }
-      if(@{$input{$_}} > @{$while}) {
-	$while = $input{$_};
-      }
-      push @fields, $_;
-    }
-  }
-  
-  while(@{$while}) {
-    @tbfields = ();
-    @values = ();
-    %pkey = ();
-    
-    $i = 0;
-    foreach $_ (@fields) {
-      $value = shift @{$input{$_}};
-      if(! @{$input{$_}}) {
-	delete $fields[$i];
-      }
-      if(($self->{dbsql_show_default} > 0) || ($value or $value eq '0') and defined($value) and !$self->{dbsql_pkey}->{$_} || ($value or $value eq '0')) {
-	push @tbfields, $_;
-	push @values, (($value or $value eq '0') ? "'$value'" : 'NULL');
-	if($self->{dbsql_pkey}->{$_}) {
-	  $pkey{$_} = "'".$value."'";
+  foreach $tbl (@{$self->{dbsql_tables}}) {
+    $fields{$tbl} = [];
+    foreach $_ (@{$self->{dbsql_fields}->{$tbl}}) {
+      if(defined($input{$_}) and $_ =~ m/^[^\]\[]+$/) {
+	if(ref($input{$_}) ne 'ARRAY') {
+	  $input{$_} = [$input{$_}]
 	}
+	if(@{$input{$_}} > @{$while}) {
+	  $while = $input{$_};
+	}
+	push @{$fields{$tbl}}, $_;
       }
-      $i ++;
-    }
-
-    if($update) {
-      $sql = $self->_dbsql_mk_update(\@tbfields, \@values, \%pkey);
-    }
-    else {
-      $sql = $self->_dbsql_mk_insert(\@tbfields, \@values);
-    }
-    
-    if($self->{debug}) {
-      print $sql, "\n";
-    }
-    if(! $self->{dbsql}->do($sql)) {
-      $self->{dbsql_sqlerr} = [$self->{dbsql}->errstr, $sql, $self->{dbsql}->err];
-      my %errconf = (
-		     ERRNUM => $self->{dbsql_sqlerr_show} & 1 ? $self->{dbsql}->err : gettext('can\'t be displayed'),
-		     ERRMSG => $self->{dbsql_sqlerr_show} & 2 ? $self->{dbsql}->errstr : gettext('can\'t be displayed'),
-		     SQLSTAT => $self->{dbsql_sqlerr_show} & 4 ? $sql : gettext('can\'t be displayed')
-		    );
-      $self->_add_to_output($self->{dbsql_sqlerr_tmpl},\%errconf);
-      return 0;
-    }
-    else {
-      $res ++;
     }
   }
+  $self->{dbsql}->begin_work;
+  while(@{$while}) {
+    @sql = ();
+    foreach $tbl (@{$self->{dbsql_tables}}) {
+      %pkey = ();
+      %values = ();
+      $i = 0;
+      foreach $_ (@{$fields{$tbl}}) {
+	$value = shift @{$input{$_}};
+	if(! @{$input{$_}}) {
+	  delete $fields{$tbl}->[$i];
+	}
+	if(($self->{dbsql_show_default} > 0) || ($value or $value eq '0') and defined($value) and !$self->{dbsql_pkey}->{$tbl}->{$_} || ($value or $value eq '0')) {
+	  ($key = $_) =~ s/^(.+)\.(.+)$/$2/;
+	  $key = $self->{dbsql}->quote_identifier($key);
+	  if($self->{dbsql_pkey}->{$tbl}->{$_}) {
+	    $pkey{$key} = $self->{dbsql}->quote($value);
+	  }
+	  $values{$key} = (($value or $value eq '0') ? $self->{dbsql}->quote($value) : $self->{dbsql}->quote('NULL'));
+	}
+	$i ++;
+      }
+      
+      if($update) {
+	push @sql, $self->_dbsql_mk_update([keys(%values)], [values(%values)], \%pkey, $tbl);
+      }
+      else {
+	push @sql, $self->_dbsql_mk_insert([keys(%values)], [values(%values)], $tbl);
+      }
+    }
+    foreach $_ (@{$self->{dbsql_extra_sql}}) {
+      $sql = $_->[0];
+      for($x=1; $x<@{$_}; $x++) {
+	$sql =~ s/(?!\\)(.)\?/$1.$values{$_->[$x]}/e;
+      }
+      $sql =~ s/\\\?/?/g;
+      push @sql, $sql;
+    }
+    foreach $sql (@sql) {
+      if($self->{debug}) {
+	print $sql, "\n";
+      }
+      if(! $self->{dbsql}->do($sql)) {
+	$self->_dbsql_sql_error($sql);
+	return 0;
+      }
+    }
+    $res ++;
+  }
+  $self->{dbsql}->commit;
   return $res;
 }
 
@@ -1248,12 +1419,9 @@ sub _dbsql_arr2psql {
 }
 
 sub _dbsql_mk_insert {
-  my $self = shift;
-  my $fields = shift;
-  my $values = shift;
-  my $table = shift || $self->{dbsql_table};
+  my ($self,$fields,$values,$table) = @_;
   if(ref($fields) eq 'ARRAY' && ref($values) eq 'ARRAY' && $table ne '') {
-    return 'INSERT INTO "' . $table . '" ('.join(', ', @{$fields}).') VALUES ('.join(', ', @{$values}).')';
+    return 'INSERT INTO ' . $self->{dbsql}->quote_identifier($table) . ' ('.join(', ', @{$fields}).') VALUES ('.join(', ', @{$values}).')';
   }
   else {
     return '';
@@ -1261,11 +1429,7 @@ sub _dbsql_mk_insert {
 }
 
 sub _dbsql_mk_update {
-  my $self = shift;
-  my $fields = shift;
-  my $values = shift;
-  my $pkey = shift;
-  my $table = shift || $self->{dbsql_table};
+  my ($self,$fields,$values,$pkey,$table) = @_;
   my $sql = '';
   my $i = 0;
 
@@ -1286,20 +1450,19 @@ sub _dbsql_mk_update {
 }
 
 sub _dbsql_makeconf {
-  my $self = shift;
-  my $info = shift;
+  my ($self,$info,$tbl) = @_;
   my %res = ();
   my $handler;
-  my $tmp;
+  my ($tmp,$var);
   if(ref($info) eq 'HASH') {
-
     #($res{TITLE} = $info->{name}) =~ s/^([a-z]{1})/uc($1)/e; does raise an endless loop
     $_ = $info->{name} and s/^([a-z]{1})/uc($1)/e and $res{TITLE} =  $_;
+    $info->{name} = $tbl . '.' . $info->{name} if(@{$self->{dbsql_tables}} > 1);
     $res{fname} = $info->{name};
     #if(($info->{dtyp} =~ s/^_//) && ($info->{default} =~ m/^\'\{(.*)\}\'$/)) { # {{''},{''}}
     $info->{default} =~ s/^'(.*)'$/$1/ if($info->{default});
     if($info->{default} && $info->{default} =~ m/^(\{.*,.*\})$/) {
-      $tmp = $self->_dbsql_parse($1);
+      ($tmp) = $self->_dbsql_parse($1);
       $res{NAME} = $self->_dbsql_parse_name($info->{name},$tmp);
       if($self->{dbsql_show_default} > 0) {
 	$res{VALUE} = $tmp;
@@ -1307,11 +1470,11 @@ sub _dbsql_makeconf {
     }
     else {
       $res{NAME} = $info->{name};
-      if(($self->{dbsql_show_default} > 0 )&& ! $self->{dbsql_pkey}->{$info->{name}}) {
+      if(($self->{dbsql_show_default} > 0 )&& ! $self->{dbsql_pkey}->{$tbl}->{$info->{name}}) {
 	$res{VALUE} = $info->{default};
       }
     }
-    $info->{dtyp} =~ s/^_//; # in postgresql arrays are marked by an leading '_'
+    $info->{dtyp} =~ s/^_//; # in postgresql arrays are marked by a leading '_'
     if(ref($self->{dbsql_dthandler}->{$info->{dtyp}}) eq 'CODE') {
       $handler = $self->{dbsql_dthandler}->{$info->{dtyp}};
     }
@@ -1319,24 +1482,29 @@ sub _dbsql_makeconf {
       $handler = $self->{dbsql_dthandler}->{default};
     }
     &$handler($self, \%res, $info);
-    if($self->{dbsql_pkey}->{$info->{name}} && $self->{dbsql_hide_pkey}) {
+    if($self->{dbsql_pkey}->{$tbl}->{$info->{name}} && $self->{dbsql_hide_pkey}) {
       $res{templ} = 'hidden';
       $res{TITLE} = '';
     }
 
-    if($info->{description} && $info->{description} =~ m/((?:(templ|[A-Z_]+)=.*;)+)/) {
-      foreach $_ (split(';',$1)) {
-	if(m/^(templ|[A-Z_]+)=(.*)$/) {
-	  $res{$1} = $self->_dbsql_parse($2);
+    if($info->{description}) {
+      while($info->{description} =~ m/\G.*?([A-Za-z_]+)\=(?:;|(.*?[^\\]{1});)/g) {
+	$var = $1;
+	if(defined($2)) {
+	  ($_ = $2) =~ s/\\;/;/g;
 	}
+	else {
+	  $_ = '';
+	}
+	($res{$var}) = $self->_dbsql_parse($_);
+	$res{$var} = '' unless(defined($res{$var}));
       }
     }
-
+    
     if(($self->{dbsql_show_default} > 0) || !$info->{default} and $info->{notnull}) {
       $res{ERROR} = ($res{ERROR} ? [$res{ERROR}] : []) unless(ref($res{ERROR}) eq 'ARRAY');
       push @{$res{ERROR}}, 'not_null';
     }
-
     if(ref($self->{dbsql_preconf}->{$info->{name}}) eq 'HASH') {
       foreach $_ (keys(%{$self->{dbsql_preconf}->{$info->{name}}})) {
 	$res{$_} = $self->{dbsql_preconf}->{$info->{name}}->{$_};
@@ -1351,32 +1519,48 @@ sub _dbsql_makeconf {
 sub _dbsql_parse {
   my ($self,$struc) = @_;
   return [$self->_dbsql_parse($1)] if($struc =~ m/^\{([^{}]*)\}$/);
-  return $struc if($struc =~ m/^[^{\,}]*$/);
+  my $struc2 = $struc;
+  while($struc2 =~ s/(\G|[^\\]{1})"(?!.*\\).*?"/$1/g){};
+  if($struc2 =~ m/^[^{\,}]*$/) {
+    while($struc =~ s/(^|[^\\]{1})"/$1/g){};
+    $struc =~ s/\\"/"/g;
+    return $struc;
+  }
   my @res;
-  if($struc =~ m/^([^{}]*\,[^{}]*)$/) {
-    push @res, '' if($struc =~ m/^,/);
-    push @res, split(',',$1) if($1);
+  if($struc2 =~ m/^([^"{}]*\,[^"{}]*)$/) {
+    $_ = $1;
+    push @res, '' if($struc =~ m/^,$/);
+    push @res, split(',',$_) if($_);
     push @res, '' if($struc =~ m/,$/);
     return @res;
   }
-  my ($off,$i,$lbr,$rbr,$sub,$add) = (0,0,0,0,0,0);
-  while($i<length($struc) and $_ = substr($struc, $i, 1)) {
+  my ($off,$i,$lbr,$rbr,$sub,$add,$quot) = (0,0,0,0,0,0,0);
+  my $tmp;
+  my $last = $_ = '';
+  while($i<length($struc)) {
+    $last = $_;
+    $_ = substr($struc, $i, 1);
+    last unless defined($_);
     $i ++;
-    ++ $lbr && $i == $off+1 ? ($add = 1) : 1 && next if($_ eq '{');
-    ++ $rbr && $i<length($struc) ? next : ($sub = 1) if($_ eq '}');
-    if((($_ eq ',' && ($sub = 1)) || $i == length($struc)) and $lbr == $rbr) {
-      push @res, $self->_dbsql_parse(substr($struc,$off+=$add,$i-$off-$sub));
-      $off=$i;
-      $sub = $add = 0;
-      next;
+    ++ $quot && $i<length($struc) ? next : 1 if($_ eq '"' and $last ne '\\');
+    unless($quot % 2) {
+      ++ $lbr && $i == $off+1 ? ($add = 1) : 1 && next if($_ eq '{');
+      ++ $rbr && $i<length($struc) ? next : ($sub = 1) if($_ eq '}');
+      if((($_ eq ',' && ($sub = 1)) || $i == length($struc)) and $lbr == $rbr) {
+	($tmp) = $self->_dbsql_parse(substr($struc,$off+=$add,$i-$off-$sub));
+	$add ? (push @res, [$tmp]) : (push @res, $tmp);
+	$off=$i;
+	$sub = $add = 0;
+	next;
+      }
     }
   }
-  return \@res;
+  return @res;
 }
 
 sub _dbsql_chk_check_sum {
   my($self,$md5hash,$val) = @_;
-  return 1 if($md5hash eq md5_hex(join($self->{dbsql_secret}, @{$val})));
+  return 1 if($md5hash eq md5_hex(join($self->{dbsql_secret}, @{$val}) . $self->{dbsql_secret}));
   return 0;
 }
 
@@ -1394,6 +1578,17 @@ sub _dbsql_parse_name {
   return $name;
 }
 
+sub _dbsql_sql_error {
+  my($self, $sql) = @_;
+  $self->{dbsql_sqlerr} = [$self->{dbsql}->errstr, $sql, $self->{dbsql}->err];
+  my %errconf = (
+		 ERRNUM => $self->{dbsql_sqlerr_show} & 1 ? $self->{dbsql}->err : gettext('can\'t be displayed'),
+		 ERRMSG => $self->{dbsql_sqlerr_show} & 2 ? $self->{dbsql}->errstr : gettext('can\'t be displayed'),
+		 SQLSTAT => $self->{dbsql_sqlerr_show} & 4 ? $sql : gettext('can\'t be displayed')
+		);
+  $self->_add_to_output($self->{dbsql_sqlerr_tmpl},\%errconf);
+}
+
 sub retarref {
   my $arr = shift;
   defined($arr) ? return [$arr] : return [] if(ref($arr) ne 'ARRAY');
